/**
 * @doc
 * @anue-package Anue.Bridge.Client
 * 這個類別實作了Web中與`<iframe/>`類型的[Bridge Server](->Anue.Bridge.AbstrctServerBridge)
 * 溝通時所使用的[Bridge Cient](Anue.Bridge.AbstractClientBridge)。
 *
 * 在使用前必須要將目標Bridge Server的對象（即iframe實體）放入建構子中進行初始化。
 *
 * ```Javascript
 * // 載入實作
 * import WebBridgeClient from 'anue-fe-sdk/web-bridge-client'
 *
 * // 取得frame實體
 * const portal = document.getElementById('protal')
 *
 * // 初始化
 * const client = new WebBridgeClient({
 *   target: protal
 * })
 *
 * // 確保已經連結
 * await client.attach()
 *
 * // 送出API請求
 * const apis = await client.request('common.handshake')
 * ```
 */
import Payload = Anue.Bridge.Payload;
import getty from '@utils/getty';
import AbstractClientBridge from '@libraries/Bridge/Client.abstract';

export default class Client extends AbstractClientBridge {
  private mRequestTable: Record<number | string, (args: any) => void> = {};
  private mPushHandlers: Record<string, Array<(args: any) => void>> = {};
  private mWindow: Window | null = null;

  protected listen = (message: MessageEvent) => {
    try {
      const payload = message.data as Payload;
      const requestId = payload.rid || -1;
      // invoke push handlers if uri matches
      // push handlers are generally for server push events
      for (const api in this.mPushHandlers) {
        if (payload.uri.startsWith(api)) {
          this.mPushHandlers[api].forEach(fn => fn(payload));
        }
      }

      // respond to request as well
      if (this.mRequestTable[requestId]) {
        this.mRequestTable[requestId](payload);
        delete this.mRequestTable[requestId];
      }
    } catch (err) {
      // console.log('Anue.ClientBridge failed to parse unknown event', err)
    }
  };

  private attachWindow = (
    target: HTMLIFrameElement | Window
  ): Promise<boolean> => {
    return new Promise(resolve => {
      window.addEventListener('message', this.listen);
      target.addEventListener('load', () => {
        this.mWindow = target instanceof Window ? target : target.contentWindow;
        resolve(true);
      });
    });
  };

  request = async (request: Anue.Bridge.RequestOptions): Promise<Payload> => {
    if (!this.mWindow) {
      return await AbstractClientBridge.NotInitializedError;
    }
    if (!request.uri) {
      throw {
        message: `invalid format request.uri can not be empty`
      };
    }
    request.rid = ++this.rid;
    this.mWindow.postMessage(JSON.stringify(request), '*');
    // @ts-ignore
    return new Promise(resolve => {
      this.mRequestTable[this.rid] = resolve;
    });
  };

  onPush = (
    apiName: Anue.Bridge.URI,
    fn: (args: Anue.Bridge.Payload) => void
  ): Client => {
    this.mPushHandlers[apiName] = this.mPushHandlers[apiName] || [];
    this.mPushHandlers[apiName].push(fn);
    return this;
  };

  attach = async (): Promise<boolean> => {
    const target = getty<HTMLIFrameElement | Window>(this.mConfig, ['target']);

    if (!target) {
      return await false;
    }
    return await this.attachWindow(target);
  };
}
