/**
 * @doc
 * @anue-library Anue.Debuggable
 * 試想一個情境，當我們遇到了一個只有在產品環境上能重現的問題時候你會怎麼進行修正？
 * 在大部分情況下由於考量到優化的問題，我們會試圖在`webpack`中將所有的log資訊拔除
 * 如此一來在產品模式中只有錯誤會拋出，其他資訊都被隱藏了。
 *
 * `Debuggable`就是在一個在不影響效能和整體JS檔案大小的前提下可以是情況開關的log模組
 * 任何繼承這個類別的class只要有任何`get`, `set`的行為都會產生log，並且當類別的
 * 成員方法被調用的時候也會紀錄下其參數和回傳結果。
 *
 * ## 原理
 *
 * 實際上`Debuggable`在應用程式第一次被載入時會依據條件來決定它要export的具體內容
 * 好比說：當`localStorage`中存在特定的entry時就export一個Debuggable類別，否則
 * 將會export一個空的類別。這樣的方式可以在Debuggable關閉的情況下對應用程式產生最
 * 小的影響。
 *
 * ## 監聽類別存取
 *
 * 只要單純的繼承`Debuggable`類別即可
 *
 * ```Javascript
 * import Debuggable from '@libraries/Debuggable'
 *
 * class MyClass extends Debuggable {
 *  // implementations
 * }
 *
 * ```
 *
 */
import Keys from '@constants/Keys';
import SharedContext from '@libraries/SharedContext';

let impl = class DebuggableEmptyImpl {};

class Debuggable {
  private static getRuntimeConfig = () => {
    let config = null;
    try {
      const shared = SharedContext.get<Anue.Debuggable.Config>(Keys.Debuggable);
      if (shared) return shared || {};
      config = JSON.parse(localStorage[Keys.Debuggable.Config]);
    } catch (err) {
      console.warn('Debuggable config not found, Debuggable disabled.');
    }
    return config;
  };

  private static readonly config: Anue.Debuggable.Config = Object.create(
    Debuggable.getRuntimeConfig()
  );
  private static consoleLogger = (type, name, prop, args, result) => {
    switch (type) {
      case 'FUNC':
        window.console.log('[FUNC] %s %s (%o) => %o', name, prop, args, result);
        break;
      default:
        window.console.log('[%s] %s %s', type, name, prop);
    }
  };
  private static log = (type, name, prop, args?, result?) => {
    const config = Debuggable.config;
    const logger =
      SharedContext.get<Anue.Debuggable.Logger>(Keys.Debuggable.Logger) ||
      Debuggable.consoleLogger;
    let shouldLog = true;
    if (config) {
      const typeMatch = config.type ? new RegExp(config.type).test(type) : true;
      const nameMatch = config.name ? new RegExp(config.name).test(name) : true;
      const propMatch = config.prop ? new RegExp(config.prop).test(prop) : true;
      shouldLog = typeMatch && nameMatch && propMatch;
    }
    if (shouldLog) {
      logger(type, name, prop, args, result);
    }
  };

  constructor() {
    return new Proxy(this, {
      construct: function(target, args) {
        //@ts-ignore
        let instance = this.apply(target, args);
        return instance;
      },
      get: function(target, prop) {
        if (typeof Reflect.get(target, prop) === 'function') {
          const org = target[prop];
          return function(...args) {
            //@ts-ignore
            let result = org.apply(this, args);
            Debuggable.log('FUNC', target.constructor.name, prop, args, result);
            return result;
          };
        }
        Debuggable.log('GET', target.constructor.name, prop);
        return prop in target ? target[prop] : undefined;
      },
      set: function(target, prop, value) {
        target[prop] = value;
        Debuggable.log('SET', target.constructor.name, prop, value);
        return true;
      }
    });
  }
}

if (
  (typeof localStorage !== 'undefined' &&
    localStorage[Keys.Debuggable.Config]) ||
  SharedContext.get(Keys.Debuggable.Config)
) {
  impl = Debuggable;
}

export default impl;
