/**
 * @anue-package Anue.Auth
 * @doc
 * Auth模組負責管理Token的交換、管理、刷新等環節。目前的架構上，由於使用者痛
 * 一使用[Cognito](https://aws.amazon.com/tw/cognito/)進行統一的身份識別因此不論是透過第三方的身份或是我們自己的
 * 認證機制（如Email）最後都得再透過API來取得Cognito ID和authorization token。
 *
 * ## 範例 (Social與Email登入)
 *
 * ```Javascript
 * import 'anue-fe-sdk'
 * import 'anue-fe-sdk/web-preset'
 * import Auth from 'anue-fe-sdk/Auth'
 * import GoogleWebProvider from 'anue-fe-sdk/GoogleWebProvider'
 * import FBWebProvider from 'anue-fe-sdk/FBWebProvider'
 * import EmailProvider from 'anue-fe-sdk/EmailProvider'
 * import Constants from 'anue-fe-sdk/Constants
 *
 * const google = new GoogleWebProvider({ apiKey, clientId })
 * const fb = new FBWebProvider({ clientId })
 * const email = new EmailProvider()
 *
 *
 * const auth = new Auth()
 *  // optional
 *  .channel(Constants.SupportedChannel.Stock)
 *  .onStateChange(ctx => {
 *    console.log('Auth state changed', ctx)
 *  })
 *
 * auth.connectProvider(google, fb, email)
 *
 *
 * ```
 */
import AbstractProvider from '@packages/Provider/AbstractProvider.abstract';
import Endpoints from '@constants/Endpoints';
import Keys, { SupportedChannel } from '@constants/Keys';
import Misc from '@constants/Misc';
import urly from '@utils/urly';
import * as Network from '@libraries/Network';
import * as Storage from '@libraries/Storage';
import getty from '@utils/getty';
import Eventy, { EventyHandler } from '@utils/eventy';
import Qable from '@utils/qable';
import { Web } from '@utils/isomorphy';
import urld from '@utils/urld';

const refreshTokenEndpoints = [
  Endpoints.v1User.email.renew(),
  Endpoints.v1User.cognitoOpenIdOrRegister()
];

class Auth {
  private mHost = Auth.getInitialURL();
  private mChannel: Anue.Auth.SupportedChannel = SupportedChannel.Unknown;
  private mIdToken: string | undefined;
  private mAccessToken: string | undefined;
  private mCognitoId: string | undefined;
  private mAuthorization: string | undefined;
  private mIsNewComer: boolean = false;
  private removeRequestInspector: (() => void) | null = null;
  private removeResponseInspector: (() => void) | null = null;
  private mConnectedProviders: Record<string, AbstractProvider> = {};
  private mIsReady: boolean = false;
  private mPrimaryProvider: AbstractProvider | null = null;
  private mProfile: Anue.Auth.UserProfile | null = null;
  private mSSOProvider: string = '';
  private mPortalUrl: string = 'https://login.cnyes.com';
  private mPostLoginProvider: string = '';
  private mSyncLock: Record<string, boolean> = {};

  event: Eventy = new Eventy();

  constructor(config: Anue.Auth.Config = {}) {
    const { channel, host } = config;

    if (channel) {
      this.mChannel = channel;
    }
    if (host) {
      this.mHost = host;
    }
  }

  host = (host: string): Auth => {
    this.mHost = host;
    return this;
  };

  channel = (channel: SupportedChannel): Auth => {
    this.mChannel = channel;
    return this;
  };

  portal = (url: string): Auth => {
    this.mPortalUrl = url;
    return this;
  };

  launchPortal = async () => {
    const state = Date.now()
      .toString()
      .replace(/\d/, s => String.fromCharCode(+s + 97));
    Storage.getDriver().setItem('anue.state', state);
    location.href = `${this.mPortalUrl}?anue_auth_state=${state}&ref=${
      this.mChannel
    }`;
  };

  ready = async (): Promise<Anue.Auth.Credentials | null> => {
    if (this.mIsReady) {
      return await this.getCredentialContext();
    }
    const value = await Storage.getDriver().getItem(
      Keys.Auth.Storage.SSOProvider
    );
    this.mSSOProvider = value;

    return new Promise<Anue.Auth.Credentials | null>(resolve => {
      const remove = this.event.on(Keys.Auth.Events.Loaded, () => {
        resolve(this.getCredentialContext());
        remove();
      });
    });
  };

  getPrimaryProvider = (): AbstractProvider | null => this.mPrimaryProvider;

  /**
   * @doc
   * @method onStateChange
   * @param fn EventyHandler<Anue.Auth.StateChangeEvent> 處理Auth狀態改變的函數，接受的參數是一個[Anue.Auth.StateChangeEvent]()物件
   * @returns [Auth](->Anue.Auth)
   * @public
   * 透過這個方法可以加入一個函數在Auth的狀態改變時被觸發，這也是應用端和Auth互動最主要的一個API。
   */
  onStateChange = (fn: EventyHandler<Anue.Auth.StateChangeEvent>): Auth => {
    this.event.on(Keys.Auth.Events.StateChange, fn);
    return this;
  };

  onError = (fn: EventyHandler<Anue.Auth.ErrorChangeEvent>) => {
    this.event.on(Keys.Auth.Events.Error, fn);
    return this;
  };

  @Web(Promise.resolve(null))
  private consumeState = async () => {
    const state = urld.get<string>('state');
    const providerCode = urld.get<string>('provider');
    const onetime = urld.get<string>('onetime');
    this.mPostLoginProvider = urld.get<string>('post_login')
      ? providerCode
      : '';
    if (state && providerCode) {
      const storage = Storage.getDriver();
      const stored = await storage.getItem('anue.state');
      console.log(
        '%c[Auth] matching state %s with stored %s provider = ',
        'color:blue;',
        state,
        stored,
        providerCode
      );
      if (state === stored && !onetime) {
        this.mSSOProvider = providerCode;
        await storage.setItem(Keys.Auth.Storage.SSOProvider, this.mSSOProvider);
        storage.setItem('anue.state', '');
      } else {
        console.log('%c[Auth] failed to match stored state', 'color:red;');
      }
    }
  };

  setSSOProvider = async (name: string) => {
    const storage = Storage.getDriver();
    await storage.setItem(Keys.Auth.Storage.SSOProvider, this.mSSOProvider);
    await storage.setItem('anue.state', '');
    this.mSSOProvider = name;
    return this;
  };

  /**
   * @doc
   * @private
   * @method requestNetworkInspector
   * >! 這個方法不為外部所引用
   * 這個方法會在認證步驟`renew`或者`consume`完成之後自動由`Auth`添加至`Network`的
   * `requestInspector`佇列之中，此後每當有任何請求時`authorization`會被自動加入在
   * 請求的header中。
   */
  private requestNetworkInspector: Anue.Network.RequestInterceptor = async (
    options: Anue.Network.RequestOptions
  ) => {
    if (this.mAuthorization) {
      options.headers = options.headers || {};
      options.headers.authorization =
        options.headers.authorization || this.mAuthorization;
    }
    return await options;
  };

  /**
   * @doc
   * @private
   * @method responseNetworkInspector
   * >! 這個方法不為外部所引用
   * 這個方法會在認證步驟`renew`或者`consume`完成之後自動由`Auth`添加至`Network`的
   * `responseInspector`佇列之中，之後當請求回應狀態代碼為`401`時，這個方法會自動呼叫
   * [Auth.renew](#renew)並且試著將原本的請求**重送**。
   */
  private responseNetworkInspector: Anue.Network.ResponseInterceptor = async (
    options: Anue.Network.RequestOptions,
    response: Anue.Network.Response
  ) => {
    const { url } = options;
    const { statusCode } = response;
    const isResendRequest = getty(options, ['extra', 'resent']);
    const isNeedRefresh =
      statusCode === 401 &&
      // do not resend refresh token requests
      refreshTokenEndpoints.reduce((p, c) => p || url.includes(c), false) ===
        false;

    if (isResendRequest || !isNeedRefresh || !this.mPrimaryProvider) {
      return await response;
    }
    console.log(
      '%c[Auth] a 401 response captured, interceptor is trying to renew the credentials ..',
      'color: red;'
    );
    // try to refresh when 401 or refresh token failed
    const credentials = await this.renew('Auth.Interceptor');
    // if the credentials successfully updated, resend original request
    // otherwise just throw the error
    if (credentials) {
      options.extra = {
        ...options.extra,
        resent: true
      };
      response = await Network.getDriver().send(options);
    }

    return await response;
  };

  /**
   * @doc
   * @method getInitialURL
   * @private
   * 這個方法會依照當前的環境返回預設的登入驗證伺服器URL，在一般瀏覽器的環境下
   * 會返回當前的`hostname`作為預設的host來發出登入等相關的API
   * 在`React Native`或者是`NodeJS`這類的環境下預設會產生一個**空字串**
   */
  private static getInitialURL = (): string => {
    return typeof location === 'undefined'
      ? ''
      : location.protocol + '//' + location.host;
  };

  /**
   * @doc
   * @public
   * @promise
   * @method renew
   * @param provider? 一個Provider的實體
   * @returns Anue.Auth.Credentials | null
   * 使用目前所儲存的使用者資訊來刷新`authorization` (呼叫我們自己提供的API時使用的token)
   * 對於`Email`來說這個行為等同於`登入`。
   */
  @Qable
  renew = async (ref?: any): Promise<Anue.Auth.Credentials | null> => {
    console.log('[Auth] renew called ref = %s', ref);
    if (!this.mPrimaryProvider) {
      console.log('[Auth] renew failed', null);
      return null;
    }
    let response, profile;
    try {
      response = await this.renewRequest();
    } catch (err) {
      // failed to refresh token, register user instead
      console.log(
        '%c[Auth] failed to refresh token, attempt to register user ..',
        'color: red'
      );
      response = await this.renewRequest(true);
    }
    console.log('[Auth] is trying to update user profile ... %o', response);
    profile = await this.mPrimaryProvider.getProfile({
      token: response.items.token,
      channel: this.mChannel,
      identityId: response.items.identityId
    });

    // user registration failed
    if (!profile) {
      this.event.emit(Keys.Auth.Events.Error, {
        source: this.mPrimaryProvider.getName(),
        detail: {
          message: 'Failed to get profile'
        }
      });
      return null;
    }
    this.mProfile = profile;

    console.log('[Auth] latest profile is %o', this.mProfile);
    return this.setCredentials(response.items).getCredentialContext();
  };

  @Qable
  private renewRequest = async (
    registration: boolean = false
  ): Promise<Anue.Auth.RefreshTokenResponse> => {
    let endpoint, body;

    if (!this.mAccessToken) {
      throw 'auth.renew cannot be called without `accessToken`';
    }

    const network = Network.getDriver();
    const renewer = getty<() => Anue.Auth.RefreshTokenResponse>(
      this.mPrimaryProvider,
      ['renewer']
    );

    let response, token;

    // use renewer defined in provider instance
    if (renewer) {
      response = await renewer();
    } else {
      const isGoogleProvider = this.mIdToken;
      body = {
        type: isGoogleProvider ? 'google' : 'facebook',
        accessToken: this.mAccessToken || '',
        idToken: this.mIdToken || '',
        channel: this.mChannel,
        refresh: registration ? '0' : '1'
      } as APIRequest.v1.User.CognitoOpenIdOrRegister;

      console.log('%c[Auth] consuming .. %o', 'color:#c19774;', body);

      endpoint = Endpoints.api.v1.user.cognitoOpenIdOrRegister();

      const url = urly(this.mHost, endpoint);
      const options = {
        method: Misc.RequestMethod.POST,
        url: url.toString(),
        body
      };
      response = await network.send<Anue.Auth.RefreshTokenResponse>(options);
    }
    token = getty(response, ['items', 'token']);
    console.log('renewed token = ', token);
    if (!token) {
      this.event.emit(Keys.Auth.Events.Error, {
        source: this.mPrimaryProvider && this.mPrimaryProvider.getName(),
        detail: response
      });
      throw {
        message: `Failed to renew authorization token, see detail for more information.`,
        detail: response
      };
    }

    return response;
  };

  private initializeProviders = async (
    providers: AbstractProvider[],
    lazy: boolean = false
  ) => {
    console.log(
      '%c⚡️[Auth] connect providers %o lazy = %o',
      'font-weight:bolder;',
      providers,
      lazy
    );

    await this.consumeState();
    const providerCode = await Storage.getDriver().getItem(
      Keys.Auth.Storage.SSOProvider
    );

    if (providerCode) {
      this.mSSOProvider = providerCode;
    }
    const promises = providers.map(provider => {
      const providerName = provider.getName();
      this.mConnectedProviders[providerName] = provider;
      if (this.mSSOProvider && providerName !== this.mSSOProvider) {
        console.log(
          '%c[Auth] drop provider [%s] as it is not sso provider',
          'color:gray;',
          providerName
        );
        return Promise.resolve({});
      }
      provider.onLoginStateChange(ctx => {
        this.syncProviderToken(provider, ctx);
      });
      console.log(
        'post login provider = %s test = %s',
        this.mPostLoginProvider,
        providerName
      );
      if (this.mPostLoginProvider === providerName) {
        console.log('[Auth] post login %s', providerName);
        return provider.init().then(() => provider.login());
      }
      return lazy ? Promise.resolve({}) : provider.init();
    });
    Promise.all(promises).then(() => {
      this.mIsReady = true;
      this.event.emit(Keys.Auth.Events.Loaded);
    });
  };

  lazyConnectProvider = (...providers: AbstractProvider[]): Auth => {
    this.initializeProviders(providers, true);
    return this;
  };

  /**
   * @doc
   * @name connectProvider
   * @param provider AbstractProvider 一個[Provider](->Anue.AbstractProvider)物件
   * 將這個provider物件的狀態同步連結到當前的Auth實體，這個行為會將[auth.consume](#consume)與該
   * provider之[login](->Anue.AbstractProvder#login)與[logout](->Anue.AbstractProvder#logout)方法進行連結。
   */
  connectProvider = (...providers: AbstractProvider[]): Auth => {
    this.initializeProviders(providers);
    return this;
  };

  private setCredentials = (data: APIResponse.v1.User.Token): Auth => {
    const { token, identityId, newcomer } = data;
    if (token && identityId && this.mPrimaryProvider) {
      this.mAuthorization = token;
      this.mCognitoId = identityId;
      this.mIsNewComer = newcomer === 1;
    }

    // add inspector for handling API authorization
    if (!this.removeRequestInspector) {
      this.removeRequestInspector = Network.addRequestInspector(
        this.requestNetworkInspector
      );
    }
    if (!this.removeResponseInspector) {
      this.removeRequestInspector = Network.addResponseInspector(
        this.responseNetworkInspector
      );
    }
    console.log(
      '%c[Auth] %cstate changed %conline',
      'font-weight:bolder;background-color: #b3ef99;',
      'background-color: #b3ef99;',
      'font-weight:bolder;background-color: #b3ef99'
    );
    this.event.emit(Keys.Auth.Events.StateChange, {
      status: 'online',
      source: this.mPrimaryProvider && this.mPrimaryProvider.getName(),
      context: this.getCredentialContext()
    });
    return this;
  };

  @Qable
  private syncProviderToken = async (
    provider: AbstractProvider,
    ctx: Anue.Provider.NullableResponse
  ) => {
    const providerName = provider.getName();
    console.log(
      '[Auth] login state change triggered by provider [%s] %o',
      providerName,
      ctx
    );
    if (this.mSSOProvider && provider.getName() !== this.mSSOProvider) {
      console.log(
        '%c[Auth] renew aborted, as %s is not dedicated SSO provider.',
        'color: gray;',
        providerName
      );
    } else if (ctx && ctx.accessToken && !this.mSyncLock[providerName]) {
      console.log('provider [%s] sync locked');
      this.mSyncLock[providerName] = true;
      this.mAccessToken = ctx.accessToken;
      this.mIdToken = ctx.idToken || '';
      this.mPrimaryProvider = provider;
      try {
        await this.renew(`Provider.${providerName}`);
      } catch (err) {
        this.mSyncLock[providerName] = false;
        throw err;
      }
      console.log('provider [%s] sync unlocked');
    } else {
      this.event.emit(Keys.Auth.Events.StateChange, {
        status: 'offline',
        source: providerName
      });
    }
  };

  @Web(null)
  private rpcCall = (prefix, ...args) => {
    console.log('[Auth] invoke rpc via dpi frame', prefix, args);
    let dpi: HTMLIFrameElement | null = document.getElementById(
      'anue-dpi'
    ) as HTMLIFrameElement;
    if (!dpi) {
      dpi = document.createElement('iframe');
      dpi.setAttribute('id', 'anue-dpi');
      dpi.style.border = 'none';
      dpi.style.position = 'fixed';
      dpi.style.left = '-9999px';
      dpi.style.top = '-9999px';
      dpi.height = '1px';
      dpi.width = '1px';
      document.body.appendChild(dpi);
    }

    dpi.addEventListener('load', () => {
      if (dpi && dpi.contentWindow) {
        console.log('post rpc event to dpi');
        dpi.contentWindow.postMessage('handshake', '*');
      }
    });
    dpi.src = `${this.mPortalUrl}/?dpi=${prefix}:${encodeURIComponent(
      args.join('.')
    )}`;
    console.log('logout from ', dpi.src);
  };

  /**
   * @doc
   * @method clearCredentials
   * 清除所有儲存的認證資訊，等同於`logout`。當這個Auth物件有使用[connectProvider](->Anue.Auth#connectProvider)
   * 連結其他[Provider](->Anue.AbstractProvider)這個方法會在連結的Provider呼叫`logout`時一併被呼叫。
   */
  clearCredentials = async () => {
    this.mAuthorization = '';
    this.mIdToken = '';
    this.mAccessToken = '';
    this.mIdToken = '';
    this.mIsNewComer = false;
    this.mCognitoId = '';

    // remove network auth inspector when credentials are cleared
    if (this.removeRequestInspector) {
      this.removeRequestInspector();
    }
    if (this.removeResponseInspector) {
      this.removeResponseInspector();
    }
    this.removeRequestInspector = null;
    this.removeResponseInspector = null;

    console.log(
      '%c[Auth] %cstate changed %coffline',
      'font-weight:bolder;background-color: #fff693;',
      'background-color: #fff693;',
      'font-weight:bolder;background-color: #fff693;'
    );

    this.rpcCall('logout', this.mChannel);
    this.event.emit(Keys.Auth.Events.StateChange, {
      status: 'offline',
      source: this.mPrimaryProvider && this.mPrimaryProvider.getName()
    });
  };

  // alias method
  logout = this.clearCredentials;

  /**
   * Get last stored credential context object
   */
  getCredentialContext = (): Anue.Auth.Credentials | null => {
    return {
      authorization: this.mAuthorization,
      idToken: this.mIdToken,
      accessToken: this.mAccessToken,
      isNewComer: this.mIsNewComer,
      cognitoId: this.mCognitoId,
      profile: this.mProfile,
      ssoProvider: this.mSSOProvider
    };
  };
}

export default Auth;
