import AbstractProvider from '@packages/Provider/AbstractProvider.abstract';
import * as Storage from '@libraries/Storage';
import Misc from '@constants/Misc';
import Endpoints from '@constants/Endpoints';
import * as Network from '@libraries/Network';
import getty from '@utils/getty';
import { obfuscate, defuscate } from '@utils/obfuscator';
import urld from '@utils/urld';

export interface IMailAndPassword {
  email: string;
  password: string;
}

export default class AnueMailProvider extends AbstractProvider {
  private static mHost =
    typeof location === 'undefined'
      ? ''
      : location.protocol + '//' + location.host;

  getName = () => 'email';

  handleStateChangeImpl() {
    this.emitStateChange();
  }

  loginImpl = async (
    args?: IMailAndPassword
  ): Promise<Anue.Provider.NullableResponse> => {
    try {
      if (!args || !args.email || !args.password) {
        throw {
          message: `AnueMailProvider can not login with improper arguments "${JSON.stringify(
            args
          )}"`
        };
      }
      this.mContext = {
        accessToken: JSON.stringify(args),
        tokenType: ''
      };
    } catch (err) {
      console.warn(err);
      return await null;
    }
    const resp = await this.renewer();

    if (resp.statusCode !== 200) {
      throw resp;
    }

    return await this.mContext;
  };

  logoutImpl = async (): Promise<boolean> => {
    this.mContext = null;
    return await true;
  };

  refresh = this.loginImpl;

  renewer = async (): Promise<Anue.Auth.RefreshTokenResponse> => {
    const body = getty<string>(this.mContext, ['accessToken']);

    if (!body) {
      throw {
        message:
          'AnueMailProvider unable to renew token as its context is empty'
      };
    }

    const resp = await Network.getDriver().send<
      APIResponse.Response<APIResponse.v1.User.Email.Renew>
    >({
      method: Misc.RequestMethod.POST,
      url: Endpoints.v1User.email.renew(),
      body: JSON.parse(body)
    });

    // store obfuscated password and email if it's valid
    if (resp.statusCode === 200) {
      await Storage.getDriver().setItem(
        `anue-mail:${AnueMailProvider.mHost}`,
        obfuscate(body)
      );
    }

    return resp;
  };

  initImpl = async (): Promise<AbstractProvider> => {
    const host = AnueMailProvider.mHost;
    const storage = Storage.getDriver();
    const urlToken = urld.get<string>('mail_access_token') || null;
    const storedToken = await storage.getItem(`anue-auth:${host}`);

    // hash token not found, try to find from storage
    if (urlToken) {
      this.mContext = {
        accessToken: defuscate(urlToken),
        tokenType: ''
      };
      console.log(
        '%c[AnueMailProvider] consumed access token from url hash %o',
        'color: #0077ff;',
        this.mContext
      );
      await this.renewer();
    } else if (storedToken) {
      try {
        const ctx = JSON.parse(storedToken);
        const provider = ctx['anue.provider'];
        const accessToken = ctx['anue.accessToken'];

        if (accessToken && provider === this.getName()) {
          this.mContext = {
            accessToken: defuscate(accessToken),
            tokenType: ''
          };
          await this.renewer();
        }
      } catch (err) {
        console.warn(err);
      }
    }
    this.emitStateChange();

    return await this;
  };

  getContext(): Anue.Provider.NullableResponse {
    return this.mContext;
  }
}
