import AbstractProvider from './AbstractProvider.abstract';
import getty from '@utils/getty';
import Keys from '@constants/Keys';

interface GAuthConfig {
  redirect_uri?: string;
  ux_mode?: 'popup' | 'redirect';
  client_id: string;
  hosted_domain?: string;
  api_key?: string;
  cookie_policy?: string;
  prompt?: string;
}

export default class GoogleProvider extends AbstractProvider {
  private mIsScriptLoaded = false;
  private loginPromises: Array<() => void> = [];
  private logoutPromises: Array<() => void> = [];
  private mAuth;
  private redirectUrl: string | null = null;
  private mClientConfig;

  static transform = <T = any>(obj): T => {
    const result = {};
    for (const prop in obj) {
      const match = prop.match(/[A-Z]?[a-z]+/g);
      if (match) {
        const key = match.join('_').toLowerCase();
        result[key] = obj[prop];
      }
    }
    return result as T;
  };

  constructor(config: Anue.Provider.GoogleProviderConfig) {
    super();
    if (!config.clientId || !config.apiKey) {
      throw {
        message:
          'Failed to initialize GoogleProvider without config.appId and config.apiKey'
      };
    }
    this.mClientConfig = config;
  }

  private getCurrentUser = () => {
    const user = getty(this, ['mAuth', 'currentUser']);
    if (user) {
      return user.get();
    }
    return null;
  };

  getName = () => 'google';

  getType = () => Keys.SupportedProviderType.Social;

  getContext = (): Anue.Provider.Response | null => {
    const currentUser = this.getCurrentUser();
    if (currentUser.isSignedIn()) {
      const ctx = currentUser.getAuthResponse(true);
      return {
        idToken: ctx.id_token,
        tokenType: ctx.token_type,
        accessToken: ctx.access_token
      };
    }
    return null;
  };

  refresh = async (): Promise<Anue.Provider.Response> => {
    if (!this.getCurrentUser()) {
      throw `cannot refresh Google tokens because the account is not logged in`;
    }
    const result = await this.mAuth.currentUser.get().reloadAuthResponse();
    console.log('GoogleWebProvider refreshed', result);
    return {
      idToken: result.id_token,
      accessToken: result.access_token,
      tokenType: result.token_type
    };
  };

  private populatePromises = (queue, ctx) => {
    while (queue.length > 0) {
      const f = queue.pop();
      if (f) {
        f(ctx);
      }
    }
  };

  handleLoginState = (loggedIn: boolean) => {
    this.mContext = this.getContext();
    let queue;
    console.log('GoogleProvider sign state in event =>', loggedIn);
    if (loggedIn) {
      queue = this.loginPromises;
    } else {
      queue = this.logoutPromises;
    }
    this.emitStateChange();
    this.populatePromises(queue, this.mContext);
  };

  initImpl = (): Promise<AbstractProvider> => {
    if (this.mIsScriptLoaded) {
      return Promise.resolve(this);
    }
    return new Promise((resolve, reject) => {
      this.mIsScriptLoaded = true;
      const script: HTMLScriptElement = document.createElement('script');
      script.src = 'https://apis.google.com/js/api:client.js';
      script.onload = () => {
        // @ts-ignore
        const gapi = window.gapi;
        gapi.load('auth2', () => {
          const config: GAuthConfig = GoogleProvider.transform<GAuthConfig>(
            this.mClientConfig
          );
          console.log('[GoogleWebProvider] initialize with config %o', config);
          if (this.redirectUrl) {
            config.ux_mode = 'redirect';
            config.redirect_uri = encodeURI(this.redirectUrl);
          }
          this.mAuth = gapi.auth2.init(config);
          // @ts-ignore
          console.log('gauth', window.gauth);
          this.mAuth.currentUser.listen(this.handleLoginState);
          if (this.mAuth.isSignedIn.get()) {
            this.mContext = this.mAuth.currentUser.get().getAuthResponse();
            this.handleLoginState(true);
          }

          resolve(this);
        });
      };
      script.onerror = reject;
      document.head.appendChild(script);
    });
  };

  logoutImpl = async (): Promise<boolean> => {
    // await this.mAuth.signOut();
    await this.mAuth.disconnect();
    if (!window.opener) {
      let dpi: HTMLIFrameElement | null = document.querySelector('#anue-dpi');
      if (!dpi) {
        dpi = document.createElement('iframe');
        dpi.setAttribute('id', 'anue-dpi');
        dpi.style.border = 'none';
        dpi.style.position = 'fixed';
        dpi.style.left = '-9999px';
        dpi.style.top = '-9999px';
        dpi.height = '1px';
        dpi.width = '1px';
        document.body.appendChild(dpi);
      }

      dpi.addEventListener('load', () => {
        if (dpi && dpi.contentWindow) {
          console.log('post logout event to dpi');
          dpi.contentWindow.postMessage('handshake', '*');
        }
      });
      dpi.src = 'https://www.google.com/accounts/Logout';
    }

    return true;
  };

  loginImpl = (config: GAuthConfig): Promise<Anue.Provider.Response | null> => {
    return this.mAuth.signIn(config);
  };
}
